#!/usr/bin/env python

"""Starts an interactive Jira session in an ipython terminal.

Script arguments support changing the server and a persistent authentication
over HTTP BASIC or Kerberos.
"""

import argparse
import configparser
import os
import sys
import webbrowser
from getpass import getpass
from urllib.parse import parse_qsl

import keyring
import requests
from oauthlib.oauth1 import SIGNATURE_RSA
from requests_oauthlib import OAuth1

from jira import JIRA, __version__

CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".jira-python", "jirashell.ini")
SENTINEL = object()


def oauth_dance(server, consumer_key, key_cert_data, print_tokens=False, verify=None):
    if verify is None:
        verify = server.startswith("https")

    # step 1: get request tokens
    oauth = OAuth1(consumer_key, signature_method=SIGNATURE_RSA, rsa_key=key_cert_data)
    r = requests.post(
        server + "/plugins/servlet/oauth/request-token", verify=verify, auth=oauth
    )
    request = dict(parse_qsl(r.text))
    request_token = request.get("oauth_token", SENTINEL)
    request_token_secret = request.get("oauth_token_secret", SENTINEL)
    if request_token is SENTINEL or request_token_secret is SENTINEL:
        problem = request.get("oauth_problem")
        if problem is not None:
            message = f"OAuth error: {problem}"
        else:
            message = " ".join(f"{key}:{value}" for key, value in request.items())
        exit(message)

    if print_tokens:
        print("Request tokens received.")
        print(f"    Request token:        {request_token}")
        print(f"    Request token secret: {request_token_secret}")

    # step 2: prompt user to validate
    auth_url = f"{server}/plugins/servlet/oauth/authorize?oauth_token={request_token}"
    if print_tokens:
        print(f"Please visit this URL to authorize the OAuth request:\n\t{auth_url}")
    else:
        webbrowser.open_new(auth_url)
        print(
            "Your browser is opening the OAuth authorization for this client session."
        )

    approved = input(
        f"Have you authorized this program to connect on your behalf to {server}? (y/n)"
    )

    if approved.lower() != "y":
        exit(
            "Abandoning OAuth dance. Your partner faceplants. The audience boos. You feel shame."
        )

    # step 3: get access tokens for validated user
    oauth = OAuth1(
        consumer_key,
        signature_method=SIGNATURE_RSA,
        rsa_key=key_cert_data,
        resource_owner_key=request_token,
        resource_owner_secret=request_token_secret,
    )
    r = requests.post(
        server + "/plugins/servlet/oauth/access-token", verify=verify, auth=oauth
    )
    access = dict(parse_qsl(r.text))

    if print_tokens:
        print("Access tokens received.")
        print(f"    Access token:        {access['oauth_token']}")
        print(f"    Access token secret: {access['oauth_token_secret']}")

    return {
        "access_token": access["oauth_token"],
        "access_token_secret": access["oauth_token_secret"],
        "consumer_key": consumer_key,
        "key_cert": key_cert_data,
    }


def process_config():
    if not os.path.exists(CONFIG_PATH):
        return {}, {}, {}, {}

    parser = configparser.ConfigParser()
    try:
        parser.read(CONFIG_PATH)
    except configparser.ParsingError as err:
        print(f"Couldn't read config file at path: {CONFIG_PATH}\n{err}")
        raise

    if parser.has_section("options"):
        options = {}
        for option, value in parser.items("options"):
            if option in ("verify", "async"):
                value = parser.getboolean("options", option)
            options[option] = value
    else:
        options = {}

    if parser.has_section("basic_auth"):
        basic_auth = dict(parser.items("basic_auth"))
    else:
        basic_auth = {}

    if parser.has_section("oauth"):
        oauth = {}
        for option, value in parser.items("oauth"):
            if option in ("oauth_dance", "print_tokens"):
                value = parser.getboolean("oauth", option)
            oauth[option] = value
    else:
        oauth = {}

    if parser.has_section("kerberos_auth"):
        kerberos_auth = {}
        for option, value in parser.items("kerberos_auth"):
            if option in ("use_kerberos"):
                value = parser.getboolean("kerberos_auth", option)
            kerberos_auth[option] = value
    else:
        kerberos_auth = {}

    return options, basic_auth, oauth, kerberos_auth


def process_command_line():
    parser = argparse.ArgumentParser(
        description="Start an interactive Jira shell with the REST API."
    )
    jira_group = parser.add_argument_group("Jira server connection options")
    jira_group.add_argument(
        "-s",
        "--server",
        help="The Jira instance to connect to, including context path.",
    )
    jira_group.add_argument(
        "-r", "--rest-path", help="The root path of the REST API to use."
    )
    jira_group.add_argument("--auth-url", help="Path to URL to auth against.")
    jira_group.add_argument(
        "-v",
        "--rest-api-version",
        help="The version of the API under the specified name.",
    )

    jira_group.add_argument(
        "--no-verify", action="store_true", help="do not verify the ssl certificate"
    )

    basic_auth_group = parser.add_argument_group("BASIC auth options")
    basic_auth_group.add_argument(
        "-u", "--username", help="The username to connect to this Jira instance with."
    )
    basic_auth_group.add_argument(
        "-p", "--password", help="The password associated with this user."
    )
    basic_auth_group.add_argument(
        "-P",
        "--prompt-for-password",
        action="store_true",
        help="Prompt for the password at the command line.",
    )

    oauth_group = parser.add_argument_group("OAuth options")
    oauth_group.add_argument(
        "-od",
        "--oauth-dance",
        action="store_true",
        help="Start a 3-legged OAuth authentication dance with Jira.",
    )
    oauth_group.add_argument("-ck", "--consumer-key", help="OAuth consumer key.")
    oauth_group.add_argument(
        "-k",
        "--key-cert",
        help="Private key to sign OAuth requests with (should be the pair of the public key\
                                   configured in the Jira application link)",
    )
    oauth_group.add_argument(
        "-pt",
        "--print-tokens",
        action="store_true",
        help="Print the negotiated OAuth tokens as they are retrieved.",
    )

    oauth_already_group = parser.add_argument_group(
        "OAuth options for already-authenticated access tokens"
    )
    oauth_already_group.add_argument(
        "-at", "--access-token", help="OAuth access token for the user."
    )
    oauth_already_group.add_argument(
        "-ats", "--access-token-secret", help="Secret for the OAuth access token."
    )

    kerberos_group = parser.add_argument_group("Kerberos options")
    kerberos_group.add_argument(
        "--use-kerberos-auth", action="store_true", help="Use kerberos auth"
    )
    kerberos_group.add_argument(
        "--mutual-authentication",
        choices=["OPTIONAL", "DISABLED"],
        help="Mutual authentication",
    )
    args = parser.parse_args()

    options = {}
    if args.server:
        options["server"] = args.server

    if args.rest_path:
        options["rest_path"] = args.rest_path

    if args.auth_url:
        options["auth_url"] = args.auth_url

    if args.rest_api_version:
        options["rest_api_version"] = args.rest_api_version

    options["verify"] = True
    if args.no_verify:
        options["verify"] = False

    if args.prompt_for_password:
        args.password = getpass()

    basic_auth = {}
    if args.username:
        basic_auth["username"] = args.username

    if args.password:
        basic_auth["password"] = args.password

    key_cert_data = None
    if args.key_cert:
        with open(args.key_cert) as key_cert_file:
            key_cert_data = key_cert_file.read()

    oauth = {}
    if args.oauth_dance:
        oauth = {
            "oauth_dance": True,
            "consumer_key": args.consumer_key,
            "key_cert": key_cert_data,
            "print_tokens": args.print_tokens,
        }
    elif (
        args.access_token
        and args.access_token_secret
        and args.consumer_key
        and args.key_cert
    ):
        oauth = {
            "access_token": args.access_token,
            "oauth_dance": False,
            "access_token_secret": args.access_token_secret,
            "consumer_key": args.consumer_key,
            "key_cert": key_cert_data,
        }

    kerberos_auth = {"use_kerberos": args.use_kerberos_auth}

    if args.mutual_authentication:
        kerberos_auth["mutual_authentication"] = args.mutual_authentication

    return options, basic_auth, oauth, kerberos_auth


def get_config():
    options, basic_auth, oauth, kerberos_auth = process_config()

    cmd_options, cmd_basic_auth, cmd_oauth, cmd_kerberos_auth = process_command_line()

    options.update(cmd_options)
    basic_auth.update(cmd_basic_auth)
    oauth.update(cmd_oauth)
    kerberos_auth.update(cmd_kerberos_auth)

    return options, basic_auth, oauth, kerberos_auth


def handle_basic_auth(auth, server):
    if auth.get("password"):
        password = auth["password"]
        if input("Would you like to remember password in OS keyring? (y/n)") == "y":
            keyring.set_password(server, auth["username"], password)
    else:
        print("Getting password from keyring...")
        password = keyring.get_password(server, auth["username"])
        assert password, "No password provided!"
    return auth["username"], password


def main():

    try:
        try:
            get_ipython
        except NameError:
            pass
        else:
            sys.exit("Running ipython inside ipython isn't supported. :(")

        options, basic_auth, oauth, kerberos_auth = get_config()

        if basic_auth:
            basic_auth = handle_basic_auth(auth=basic_auth, server=options["server"])

        if oauth.get("oauth_dance") is True:
            oauth = oauth_dance(
                options["server"],
                oauth["consumer_key"],
                oauth["key_cert"],
                oauth["print_tokens"],
                options["verify"],
            )
        elif not all(
            (
                oauth.get("access_token"),
                oauth.get("access_token_secret"),
                oauth.get("consumer_key"),
                oauth.get("key_cert"),
            )
        ):
            oauth = None

        use_kerberos = kerberos_auth.get("use_kerberos", False)
        del kerberos_auth["use_kerberos"]

        jira = JIRA(
            options=options,
            basic_auth=basic_auth,
            kerberos=use_kerberos,
            kerberos_options=kerberos_auth,
            oauth=oauth,
        )

        import IPython

        # The top-level `frontend` package has been deprecated since IPython 1.0.
        if IPython.version_info[0] >= 1:
            from IPython.terminal.embed import InteractiveShellEmbed
        else:
            from IPython.frontend.terminal.embed import InteractiveShellEmbed

        ip_shell = InteractiveShellEmbed(
            banner1="<Jira Shell " + __version__ + " (" + jira.server_url + ")>"
        )
        ip_shell("*** Jira shell active; client is in 'jira'. Press Ctrl-D to exit.")
    except Exception as e:
        print(e, file=sys.stderr)
        return 2


if __name__ == "__main__":
    status = main()
    sys.exit(status)



import json
import logging
import random
import time
from typing import Callable, Optional, Union, cast

from requests import Response, Session
from requests.exceptions import ConnectionError

from jira.exceptions import JIRAError

logging.getLogger("jira").addHandler(logging.NullHandler())


def raise_on_error(r: Optional[Response], verb="???", **kwargs):
    """Handle errors from a Jira Request

    Args:
        r (Optional[Response]): Response from Jira request
        verb (Optional[str]): Request type, e.g. POST. Defaults to "???".

    Raises:
        JIRAError: If Response is None
        JIRAError: for unhandled 400 status codes.
        JIRAError: for unhandled 200 status codes.
    """
    request = kwargs.get("request", None)
    # headers = kwargs.get('headers', None)

    if r is None:
        raise JIRAError(None, **kwargs)

    if r.status_code >= 400:
        error = ""
        if r.status_code == 403 and "x-authentication-denied-reason" in r.headers:
            error = r.headers["x-authentication-denied-reason"]
        elif r.text:
            try:
                response = json.loads(r.text)
                if "message" in response:
                    # Jira 5.1 errors
                    error = response["message"]
                elif "errorMessages" in response and len(response["errorMessages"]) > 0:
                    # Jira 5.0.x error messages sometimes come wrapped in this array
                    # Sometimes this is present but empty
                    errorMessages = response["errorMessages"]
                    if isinstance(errorMessages, (list, tuple)):
                        error = errorMessages[0]
                    else:
                        error = errorMessages
                # Catching only 'errors' that are dict. See https://github.com/pycontribs/jira/issues/350
                elif (
                    "errors" in response
                    and len(response["errors"]) > 0
                    and isinstance(response["errors"], dict)
                ):
                    # Jira 6.x error messages are found in this array.
                    error_list = response["errors"].values()
                    error = ", ".join(error_list)
                else:
                    error = r.text
            except ValueError:
                error = r.text
        raise JIRAError(
            error,
            status_code=r.status_code,
            url=r.url,
            request=request,
            response=r,
            **kwargs,
        )
    # for debugging weird errors on CI
    if r.status_code not in [200, 201, 202, 204]:
        raise JIRAError(
            status_code=r.status_code, request=request, response=r, **kwargs
        )
    # testing for the bug exposed on
    # https://answers.atlassian.com/questions/11457054/answers/11975162
    if (
        r.status_code == 200
        and len(r.content) == 0
        and "X-Seraph-LoginReason" in r.headers
        and "AUTHENTICATED_FAILED" in r.headers["X-Seraph-LoginReason"]
    ):
        pass


class ResilientSession(Session):
    """This class is supposed to retry requests that do return temporary errors.

    At this moment it supports: 502, 503, 504
    """

    def __init__(self, timeout=None):
        self.max_retries = 3
        self.max_retry_delay = 60
        self.timeout = timeout
        super().__init__()

        # Indicate our preference for JSON to avoid https://bitbucket.org/bspeakmon/jira-python/issue/46 and https://jira.atlassian.com/browse/JRA-38551
        self.headers.update({"Accept": "application/json,*.*;q=0.9"})

    def __recoverable(
        self,
        response: Optional[Union[ConnectionError, Response]],
        url: str,
        request,
        counter: int = 1,
    ):
        msg = str(response)
        if isinstance(response, ConnectionError):
            logging.warning(
                f"Got ConnectionError [{response}] errno:{response.errno} on {request} {url}\n{vars(response)}\n{response.__dict__}"
            )
        if isinstance(response, Response):
            if response.status_code in [502, 503, 504, 401]:
                # 401 UNAUTHORIZED still randomly returned by Atlassian Cloud as of 2017-01-16
                msg = f"{response.status_code} {response.reason}"
                # 2019-07-25: Disabled recovery for codes above^
                return False
            elif not (
                response.status_code == 200
                and len(response.content) == 0
                and "X-Seraph-LoginReason" in response.headers
                and "AUTHENTICATED_FAILED" in response.headers["X-Seraph-LoginReason"]
            ):
                return False
            else:
                msg = "Atlassian's bug https://jira.atlassian.com/browse/JRA-41559"

        # Exponential backoff with full jitter.
        delay = min(self.max_retry_delay, 10 * 2 ** counter) * random.random()
        logging.warning(
            "Got recoverable error from %s %s, will retry [%s/%s] in %ss. Err: %s"
            % (request, url, counter, self.max_retries, delay, msg)
        )
        if isinstance(response, Response):
            logging.debug("response.headers: %s", response.headers)
            logging.debug("response.body: %s", response.content)
        time.sleep(delay)
        return True

    def __verb(
        self, verb: str, url: str, retry_data: Callable = None, **kwargs
    ) -> Response:

        d = self.headers.copy()
        d.update(kwargs.get("headers", {}))
        kwargs["headers"] = d

        # if we pass a dictionary as the 'data' we assume we want to send json
        # data
        data = kwargs.get("data", {})
        if isinstance(data, dict):
            data = json.dumps(data)

        retry_number = 0
        exception = None
        response = None
        while retry_number <= self.max_retries:
            response = None
            exception = None
            try:
                method = getattr(super(), verb.lower())
                response = method(url, timeout=self.timeout, **kwargs)
                if response.status_code >= 200 and response.status_code <= 299:
                    return response
            except ConnectionError as e:
                logging.warning(f"{e} while doing {verb.upper()} {url}")

                exception = e
            retry_number += 1

            if retry_number <= self.max_retries:
                response_or_exception = response if response is not None else exception
                if self.__recoverable(
                    response_or_exception, url, verb.upper(), retry_number
                ):
                    if retry_data:
                        # if data is a stream, we cannot just read again from it,
                        # retry_data() will give us a new stream with the data
                        kwargs["data"] = retry_data()
                    continue
                else:
                    break

        if exception is not None:
            raise exception
        raise_on_error(response, verb=verb, **kwargs)
        # after raise_on_error, only Response objects are allowed through
        response = cast(Response, response)  # tell mypy only Response-like are here
        return response

    def get(self, url: Union[str, bytes], **kwargs) -> Response:  # type: ignore
        return self.__verb("GET", str(url), **kwargs)

    def post(self, url: Union[str, bytes], data=None, json=None, **kwargs) -> Response:  # type: ignore
        return self.__verb("POST", str(url), data=data, json=json, **kwargs)

    def put(self, url: Union[str, bytes], data=None, **kwargs) -> Response:  # type: ignore
        return self.__verb("PUT", str(url), data=data, **kwargs)

    def delete(self, url: Union[str, bytes], **kwargs) -> Response:  # type: ignore
        return self.__verb("DELETE", str(url), **kwargs)

    def head(self, url: Union[str, bytes], **kwargs) -> Response:  # type: ignore
        return self.__verb("HEAD", str(url), **kwargs)

    def patch(self, url: Union[str, bytes], data=None, **kwargs) -> Response:  # type: ignore
        return self.__verb("PATCH", str(url), data=data, **kwargs)

    def options(self, url: Union[str, bytes], **kwargs) -> Response:  # type: ignore
        return self.__verb("OPTIONS", str(url), **kwargs)



"""
This module implements the Resource classes that translate JSON from Jira REST resources
into usable objects.
"""

import json
import logging
import re
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union, cast

from requests import Response
from requests.structures import CaseInsensitiveDict

from jira.resilientsession import ResilientSession
from jira.utils import json_loads, threaded_requests

if TYPE_CHECKING:
    from jira.client import JIRA

    AnyLike = Any
else:

    class AnyLike:
        """Dummy subclass of base object class for when type checker is not running."""

        pass


__all__ = (
    "Resource",
    "Issue",
    "Comment",
    "Project",
    "Attachment",
    "Component",
    "Dashboard",
    "Filter",
    "Votes",
    "PermissionScheme",
    "Watchers",
    "Worklog",
    "IssueLink",
    "IssueLinkType",
    "IssueType",
    "Priority",
    "Version",
    "Role",
    "Resolution",
    "SecurityLevel",
    "Status",
    "User",
    "Group",
    "CustomFieldOption",
    "RemoteLink",
    "Customer",
    "ServiceDesk",
    "RequestType",
    "resource_class_map",
)

logging.getLogger("jira").addHandler(logging.NullHandler())


def get_error_list(r: Response) -> List[str]:
    error_list = []
    if r.status_code >= 400:
        if r.status_code == 403 and "x-authentication-denied-reason" in r.headers:
            error_list = [r.headers["x-authentication-denied-reason"]]
        elif r.text:
            try:
                response: Dict[str, Any] = json_loads(r)
                if "message" in response:
                    # Jira 5.1 errors
                    error_list = [response["message"]]
                elif "errorMessages" in response and len(response["errorMessages"]) > 0:
                    # Jira 5.0.x error messages sometimes come wrapped in this array
                    # Sometimes this is present but empty
                    errorMessages = response["errorMessages"]
                    if isinstance(errorMessages, (list, tuple)):
                        error_list = list(errorMessages)
                    else:
                        error_list = [errorMessages]
                elif "errors" in response and len(response["errors"]) > 0:
                    # Jira 6.x error messages are found in this array.
                    error_list = response["errors"].values()
                else:
                    error_list = [r.text]
            except ValueError:
                error_list = [r.text]
    return error_list


class Resource:
    """Models a URL-addressable resource in the Jira REST API.

    All Resource objects provide the following:
    ``find()`` -- get a resource from the server and load it into the current object
    (though clients should use the methods in the JIRA class instead of this method directly)
    ``update()`` -- changes the value of this resource on the server and returns a new resource object for it
    ``delete()`` -- deletes this resource from the server
    ``self`` -- the URL of this resource on the server
    ``raw`` -- dict of properties parsed out of the JSON response from the server

    Subclasses will implement ``update()`` and ``delete()`` as appropriate for the specific resource.

    All Resources have a resource path of the form:

    * ``issue``
    * ``project/{0}``
    * ``issue/{0}/votes``
    * ``issue/{0}/comment/{1}``

    where the bracketed numerals are placeholders for ID values that are filled in from the
    ``ids`` parameter to ``find()``.
    """

    JIRA_BASE_URL = "{server}/rest/{rest_path}/{rest_api_version}/{path}"

    # A prioritized list of the keys in self.raw most likely to contain a human
    # readable name or identifier, or that offer other key information.
    _READABLE_IDS = (
        "displayName",
        "key",
        "name",
        "accountId",
        "filename",
        "value",
        "scope",
        "votes",
        "id",
        "mimeType",
        "closed",
    )

    # A list of properties that should uniquely identify a Resource object
    # Each of these properties should be hashable, usually strings
    _HASH_IDS = (
        "self",
        "type",
        "key",
        "id",
        "name",
    )

    def __init__(
        self,
        resource: str,
        options: Dict[str, Any],
        session: ResilientSession,
        base_url: str = JIRA_BASE_URL,
    ):
        """Initializes a generic resource.

        Args:
            resource (str): The name of the resource.
            options (Dict[str,str]): Options for the new resource
            session (ResilientSession): Session used for the resource.
            base_url (Optional[str]): The Base Jira url.

        """
        self._resource = resource
        self._options = options
        self._session = session
        self._base_url = base_url

        # Explicitly define as None so we know when a resource has actually
        # been loaded
        self.raw: Optional[Dict[str, Any]] = None

    def __str__(self) -> str:
        """Return the first value we find that is likely to be human readable.

        Returns:
            str
        """
        if self.raw:
            for name in self._READABLE_IDS:
                if name in self.raw:
                    pretty_name = str(self.raw[name])
                    # Include any child to support nested select fields.
                    if hasattr(self, "child"):
                        pretty_name += " - " + str(self.child)
                    return pretty_name

        # If all else fails, use repr to make sure we get something.
        return repr(self)

    def __repr__(self) -> str:
        """Identify the class and include any and all relevant values.

        Returns:
            str
        """
        names: List[str] = []
        if self.raw:
            for name in self._READABLE_IDS:
                if name in self.raw:
                    names.append(name + "=" + repr(self.raw[name]))
        if not names:
            return f"<JIRA {self.__class__.__name__} at {id(self)}>"
        return f"<JIRA {self.__class__.__name__}: {', '.join(names)}>"

    def __getattr__(self, item: str) -> Any:
        """Allow access of attributes via names.

        Args:
            item (str): Attribute Name

        Raises:
            AttributeError: When attribute does not exist.

        Returns:
            Any: Attribute value.
        """
        try:
            return self[item]  # type: ignore
        except Exception as e:
            if hasattr(self, "raw") and self.raw is not None and item in self.raw:
                return self.raw[item]
            else:
                raise AttributeError(
                    f"{self.__class__!r} object has no attribute {item!r} ({e})"
                )

    def __getstate__(self) -> Dict[str, Any]:
        """Pickling the resource."""
        return vars(self)

    def __setstate__(self, raw_pickled: Dict[str, Any]):
        """Unpickling of the resource"""
        # https://stackoverflow.com/a/50888571/7724187
        vars(self).update(raw_pickled)

    def __hash__(self) -> int:
        """Hash calculation.

        We try to find unique identifier like properties
        to form our hash object.
        Technically 'self', if present, is the unique URL to the object,
        and should be sufficient to generate a unique hash.
        """
        hash_list = []
        for a in self._HASH_IDS:
            if hasattr(self, a):
                hash_list.append(getattr(self, a))

        if hash_list:
            return hash(tuple(hash_list))
        else:
            raise TypeError(f"'{self.__class__}' is not hashable")

    def __eq__(self, other: Any) -> bool:
        """Default equality test.

        Checks the types look about right and that the relevant
        attributes that uniquely identify a resource are equal.
        """
        return isinstance(other, self.__class__) and all(
            [
                getattr(self, a) == getattr(other, a)
                for a in self._HASH_IDS
                if hasattr(self, a)
            ]
        )

    def find(
        self,
        id: Union[Tuple[str, str], int, str],
        params: Optional[Dict[str, str]] = None,
    ):
        """Finds a resource based on the input parameters.

        Args:
            id (Union[Tuple[str, str], int, str]): id
            params (Optional[Dict[str, str]]): params

        """

        if params is None:
            params = {}

        if isinstance(id, tuple):
            path = self._resource.format(*id)
        else:
            path = self._resource.format(id)
        url = self._get_url(path)
        self._load(url, params=params)

    def _get_url(self, path: str) -> str:
        """Gets the url for the specified path.

        Args:
            path (str): str

        Returns:
            str
        """
        options = self._options.copy()
        options.update({"path": path})
        return self._base_url.format(**options)

    def update(
        self,
        fields: Optional[Dict[str, Any]] = None,
        async_: Optional[bool] = None,
        jira: "JIRA" = None,
        notify: bool = True,
        **kwargs: Any,
    ):
        """Update this resource on the server.

        Keyword arguments are marshalled into a dict before being sent. If this
        resource doesn't support ``PUT``, a :py:exc:`.JIRAError` will be raised; subclasses that specialize this method
        will only raise errors in case of user error.

        Args:
            fields (Optional[Dict[str, Any]]): Fields which should be updated for the object.
            async_ (bool): If true the request will be added to the queue so it can be executed later using async_run()
            jira (jira.client.JIRA): Instance of Jira Client
            notify (bool): Whether or not to notify users about the update. (Default: True)
            kwargs (Any): extra arguments to the PUT request.
        """
        if async_ is None:
            async_: bool = self._options["async"]  # type: ignore # redefinition

        data = {}
        if fields is not None:
            data.update(fields)
        data.update(kwargs)

        if not notify:
            querystring = "?notifyUsers=false"
        else:
            querystring = ""

        r = self._session.put(self.self + querystring, data=json.dumps(data))
        if "autofix" in self._options and r.status_code == 400:
            user = None
            error_list = get_error_list(r)
            logging.error(error_list)
            if "The reporter specified is not a user." in error_list:
                if "reporter" not in data["fields"]:
                    logging.warning(
                        "autofix: setting reporter to '%s' and retrying the update."
                        % self._options["autofix"]
                    )
                    data["fields"]["reporter"] = {"name": self._options["autofix"]}

            if "Issues must be assigned." in error_list:
                if "assignee" not in data["fields"]:
                    logging.warning(
                        "autofix: setting assignee to '%s' for %s and retrying the update."
                        % (self._options["autofix"], self.key)
                    )
                    data["fields"]["assignee"] = {"name": self._options["autofix"]}
                    # for some reason the above approach fails on Jira 5.2.11
                    # so we need to change the assignee before

            if (
                "Issue type is a sub-task but parent issue key or id not specified."
                in error_list
            ):
                logging.warning(
                    "autofix: trying to fix sub-task without parent by converting to it to bug"
                )
                data["fields"]["issuetype"] = {"name": "Bug"}
            if (
                "The summary is invalid because it contains newline characters."
                in error_list
            ):
                logging.warning("autofix: trying to fix newline in summary")
                data["fields"]["summary"] = self.fields.summary.replace("/n", "")
            for error in error_list:
                if re.search(
                    r"^User '(.*)' was not found in the system\.", error, re.U
                ):
                    m = re.search(
                        r"^User '(.*)' was not found in the system\.", error, re.U
                    )
                    if m:
                        user = m.groups()[0]
                    else:
                        raise NotImplementedError()
                if re.search(r"^User '(.*)' does not exist\.", error):
                    m = re.search(r"^User '(.*)' does not exist\.", error)
                    if m:
                        user = m.groups()[0]
                    else:
                        raise NotImplementedError()

            if user and jira:
                logging.warning(
                    "Trying to add missing orphan user '%s' in order to complete the previous failed operation."
                    % user
                )
                jira.add_user(user, "noreply@example.com", 10100, active=False)
                # if 'assignee' not in data['fields']:
                #    logging.warning("autofix: setting assignee to '%s' and retrying the update." % self._options['autofix'])
                #    data['fields']['assignee'] = {'name': self._options['autofix']}
            # EXPERIMENTAL --->
            if async_:  # FIXME: no async
                if not hasattr(self._session, "_async_jobs"):
                    self._session._async_jobs = set()  # type: ignore
                self._session._async_jobs.add(  # type: ignore
                    threaded_requests.put(  # type: ignore
                        self.self, data=json.dumps(data)
                    )
                )
            else:
                r = self._session.put(self.self, data=json.dumps(data))

        time.sleep(self._options["delay_reload"])
        self._load(self.self)

    def delete(self, params: Optional[Dict[str, Any]] = None) -> Optional[Response]:
        """Delete this resource from the server, passing the specified query parameters.

        If this resource doesn't support ``DELETE``, a :py:exc:`.JIRAError`
        will be raised; subclasses that specialize this method will only raise errors
        in case of user error.

        Args:
            params: Parameters for the delete request.

        Returns:
            Optional[Response]: Returns None if async
        """
        if self._options["async"]:
            # FIXME: mypy doesn't think this should work
            if not hasattr(self._session, "_async_jobs"):
                self._session._async_jobs = set()  # type: ignore
            self._session._async_jobs.add(  # type: ignore
                threaded_requests.delete(url=self.self, params=params)  # type: ignore
            )
            return None
        else:
            return self._session.delete(url=self.self, params=params)

    def _load(
        self,
        url: str,
        headers=CaseInsensitiveDict(),
        params: Optional[Dict[str, str]] = None,
        path: Optional[str] = None,
    ):
        """Load a resource.

        Args:
            url (str): url
            headers (Optional[CaseInsensitiveDict]): headers. Defaults to CaseInsensitiveDict().
            params (Optional[Dict[str,str]]): params to get request. Defaults to None.
            path (Optional[str]): field to get. Defaults to None.

        Raises:
            ValueError: If json cannot be loaded
        """
        r = self._session.get(url, headers=headers, params=params)
        try:
            j = json_loads(r)
        except ValueError as e:
            logging.error(f"{e}:\n{r.text}")
            raise e
        if path:
            j = j[path]
        self._parse_raw(j)

    def _parse_raw(self, raw: Dict[str, Any]):
        """Parse a raw dictionary to create a resource.

        Args:
            raw (Dict[str, Any])
        """
        self.raw = raw
        if not raw:
            raise NotImplementedError(f"We cannot instantiate empty resources: {raw}")
        dict2resource(raw, self, self._options, self._session)

    def _default_headers(self, user_headers):
        # result = dict(user_headers)
        # result['accept'] = 'application/json'
        return CaseInsensitiveDict(
            self._options["headers"].items() + user_headers.items()
        )


class Attachment(Resource):
    """An issue attachment."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "attachment/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def get(self):
        """Return the file content as a string."""
        r = self._session.get(self.content, headers={"Accept": "*/*"})
        return r.content

    def iter_content(self, chunk_size=1024):
        """Return the file content as an iterable stream."""
        r = self._session.get(self.content, stream=True)
        return r.iter_content(chunk_size)


class Component(Resource):
    """A project component."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "component/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def delete(self, moveIssuesTo: Optional[str] = None):  # type: ignore[override]
        """Delete this component from the server.

        Args:
            moveIssuesTo: the name of the component to which to move any issues this component is applied
        """
        params = {}
        if moveIssuesTo is not None:
            params["moveIssuesTo"] = moveIssuesTo

        super().delete(params)


class CustomFieldOption(Resource):
    """An existing option for a custom issue field."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "customFieldOption/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Dashboard(Resource):
    """A Jira dashboard."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "dashboard/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Filter(Resource):
    """An issue navigator filter."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "filter/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Issue(Resource):
    """A Jira issue."""

    class _IssueFields(AnyLike):
        class _Comment:
            def __init__(self) -> None:
                self.comments: List[Comment] = []

        class _Worklog:
            def __init__(self) -> None:
                self.worklogs: List[Worklog] = []

        def __init__(self):
            self.assignee: Optional[UnknownResource] = None
            self.attachment: List[Attachment] = []
            self.comment = self._Comment()
            self.created: str
            self.description: Optional[str] = None
            self.duedate: Optional[str] = None
            self.issuelinks: List[IssueLink] = []
            self.issuetype: IssueType
            self.labels: List[str] = []
            self.priority: Priority
            self.project: Project
            self.reporter: UnknownResource
            self.resolution: Optional[Resolution] = None
            self.security: Optional[SecurityLevel] = None
            self.status: Status
            self.statuscategorychangedate: Optional[str] = None
            self.summary: str
            self.timetracking: TimeTracking
            self.versions: List[Version] = []
            self.votes: Votes
            self.watchers: Watchers
            self.worklog = self._Worklog()

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issue/{0}", options, session)

        self.fields: Issue._IssueFields
        self.id: str
        self.key: str
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def update(  # type: ignore[override] # incompatible supertype ignored
        self,
        fields: Dict[str, Any] = None,
        update: Dict[str, Any] = None,
        async_: bool = None,
        jira: "JIRA" = None,
        notify: bool = True,
        **fieldargs,
    ):
        """Update this issue on the server.

        Each keyword argument (other than the predefined ones) is treated as a field name and the argument's value
        is treated as the intended value for that field -- if the fields argument is used, all other keyword arguments
        will be ignored.

        Jira projects may contain many different issue types. Some issue screens have different requirements for
        fields in an issue. This information is available through the :py:meth:`.JIRA.editmeta` method. Further examples
        are available here: https://developer.atlassian.com/display/JIRADEV/JIRA+REST+API+Example+-+Edit+issues

        Args:
            fields (Dict[str,Any]): a dict containing field names and the values to use
            update (Dict[str,Any]): a dict containing update operations to apply
            notify (bool): query parameter notifyUsers. If true send the email with notification that the issue was updated
              to users that watch it. Admin or project admin permissions are required to disable the notification.
            jira (Optional[jira.client.JIRA]): JIRA instance.
            fieldargs (dict): keyword arguments will generally be merged into fields, except lists,
              which will be merged into updates

        """
        data = {}
        if fields is not None:
            fields_dict = fields
        else:
            fields_dict = {}
        data["fields"] = fields_dict
        if update is not None:
            update_dict = update
        else:
            update_dict = {}
        data["update"] = update_dict
        for field in sorted(fieldargs.keys()):
            value = fieldargs[field]
            # apply some heuristics to make certain changes easier
            if isinstance(value, str):
                if field == "assignee" or field == "reporter":
                    fields_dict[field] = {"name": value}
                elif field == "comment":
                    if "comment" not in update_dict:
                        update_dict["comment"] = []
                    update_dict["comment"].append({"add": {"body": value}})
                else:
                    fields_dict[field] = value
            elif isinstance(value, list):
                if field not in update_dict:
                    update_dict[field] = []
                update_dict[field].extend(value)
            else:
                fields_dict[field] = value

        super().update(async_=async_, jira=jira, notify=notify, fields=data)

    def add_field_value(self, field: str, value: str):
        """Add a value to a field that supports multiple values, without resetting the existing values.

        This should work with: labels, multiple checkbox lists, multiple select

        Args:
            field (str): The field name
            value (str): The field's value

        """
        super().update(fields={"update": {field: [{"add": value}]}})

    def delete(self, deleteSubtasks=False):
        """Delete this issue from the server.

        Args:
            deleteSubtasks (bool): if the issue has subtasks, this argument must be set to true for the call to succeed.

        """
        super().delete(params={"deleteSubtasks": deleteSubtasks})

    def permalink(self):
        """Get the URL of the issue, the browsable one not the REST one.

        Returns:
            str: URL of the issue
        """
        return f"{self._options['server']}/browse/{self.key}"


class Comment(Resource):
    """An issue comment."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issue/{0}/comment/{1}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def update(self, fields=None, async_=None, jira=None, body="", visibility=None):
        """Update a comment"""
        data = {}
        if body:
            data["body"] = body
        if visibility:
            data["visibility"] = visibility
        super().update(data)


class RemoteLink(Resource):
    """A link to a remote application from an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issue/{0}/remotelink/{1}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def update(self, object, globalId=None, application=None, relationship=None):
        """Update a RemoteLink. 'object' is required.

        For definitions of the allowable fields for 'object' and the keyword arguments 'globalId', 'application' and
        'relationship', see https://developer.atlassian.com/display/JIRADEV/JIRA+REST+API+for+Remote+Issue+Links.

        Args:
            object: the link details to add (see the above link for details)
            globalId: unique ID for the link (see the above link for details)
            application: application information for the link (see the above link for details)
            relationship: relationship description for the link (see the above link for details)
        """
        data = {"object": object}
        if globalId is not None:
            data["globalId"] = globalId
        if application is not None:
            data["application"] = application
        if relationship is not None:
            data["relationship"] = relationship

        super().update(**data)


class Votes(Resource):
    """Vote information on an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issue/{0}/votes", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class PermissionScheme(Resource):
    """Permissionscheme information on an project."""

    def __init__(self, options, session, raw=None):
        Resource.__init__(
            self, "project/{0}/permissionscheme?expand=user", options, session
        )
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Watchers(Resource):
    """Watcher information on an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issue/{0}/watchers", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def delete(self, username):
        """Remove the specified user from the watchers list."""
        super().delete(params={"username": username})


class TimeTracking(Resource):
    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issue/{0}/worklog/{1}", options, session)
        self.remainingEstimate = None
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Worklog(Resource):
    """Worklog on an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issue/{0}/worklog/{1}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def delete(  # type: ignore[override]
        self, adjustEstimate: Optional[str] = None, newEstimate=None, increaseBy=None
    ):
        """Delete this worklog entry from its associated issue.

        Args:
            adjustEstimate: one of ``new``, ``leave``, ``manual`` or ``auto``.
              ``auto`` is the default and adjusts the estimate automatically.
              ``leave`` leaves the estimate unchanged by this deletion.
            newEstimate: combined with ``adjustEstimate=new``, set the estimate to this value
            increaseBy: combined with ``adjustEstimate=manual``, increase the remaining estimate by this amount
        """
        params = {}
        if adjustEstimate is not None:
            params["adjustEstimate"] = adjustEstimate
        if newEstimate is not None:
            params["newEstimate"] = newEstimate
        if increaseBy is not None:
            params["increaseBy"] = increaseBy

        super().delete(params)


class IssueLink(Resource):
    """Link between two issues."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issueLink/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class IssueLinkType(Resource):
    """Type of link between two issues."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issueLinkType/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class IssueType(Resource):
    """Type of an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "issuetype/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Priority(Resource):
    """Priority that can be set on an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "priority/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Project(Resource):
    """A Jira project."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "project/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Role(Resource):
    """A role inside a project."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "project/{0}/role/{1}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def update(  # type: ignore[override]
        self,
        users: Union[str, List, Tuple] = None,
        groups: Union[str, List, Tuple] = None,
    ):
        """Add the specified users or groups to this project role. One of ``users`` or ``groups`` must be specified.

        Args:
            users (Optional[Union[str,List,Tuple]]): a user or users to add to the role
            groups (Optional[Union[str,List,Tuple]]): a group or groups to add to the role
        """

        if users is not None and isinstance(users, str):
            users = (users,)
        if groups is not None and isinstance(groups, str):
            groups = (groups,)

        data = {
            "id": self.id,
            "categorisedActors": {
                "atlassian-user-role-actor": users,
                "atlassian-group-role-actor": groups,
            },
        }

        super().update(**data)

    def add_user(
        self,
        users: Union[str, List, Tuple] = None,
        groups: Union[str, List, Tuple] = None,
    ):
        """Add the specified users or groups to this project role.

        One of ``users`` or ``groups`` must be specified.

        Args:
            users (Optional[Union[str,List,Tuple]]): a user or users to add to the role
            groups (Optional[Union[str,List,Tuple]]): a group or groups to add to the role
        """

        if users is not None and isinstance(users, str):
            users = (users,)
        if groups is not None and isinstance(groups, str):
            groups = (groups,)

        data = {"user": users, "group": groups}
        self._session.post(self.self, data=json.dumps(data))


class Resolution(Resource):
    """A resolution for an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "resolution/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class SecurityLevel(Resource):
    """A security level for an issue or project."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "securitylevel/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Status(Resource):
    """Status for an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "status/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class StatusCategory(Resource):
    """StatusCategory for an issue."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "statuscategory/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class User(Resource):
    """A Jira user."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
        *,
        _query_param: str = "username",
    ):
        # Handle self-hosted Jira and Jira Cloud differently
        if raw and "accountId" in raw["self"]:
            _query_param = "accountId"

        Resource.__init__(self, f"user?{_query_param}" + "={0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Group(Resource):
    """A Jira user group."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "group?groupname={0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Version(Resource):
    """A version of a project."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "version/{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)

    def delete(self, moveFixIssuesTo=None, moveAffectedIssuesTo=None):
        """
        Delete this project version from the server.

        If neither of the arguments are specified, the version is removed from all
        issues it is attached to.

        Args:
            moveFixIssuesTo: in issues for which this version is a fix
              version, add this argument version to the fix version list
            moveAffectedIssuesTo: in issues for which this version is an
              affected version, add this argument version to the affected version list
        """

        params = {}
        if moveFixIssuesTo is not None:
            params["moveFixIssuesTo"] = moveFixIssuesTo
        if moveAffectedIssuesTo is not None:
            params["moveAffectedIssuesTo"] = moveAffectedIssuesTo

        return super().delete(params)

    def update(self, **kwargs):
        """
        Update this project version from the server. It is prior used to archive versions.

        Refer to Atlassian REST API `documentation`_.

        .. _documentation: https://developer.atlassian.com/cloud/jira/platform/rest/v2/api-group-project-versions/#api-rest-api-2-version-id-put

        :Example:

            .. code-block:: python

                >> version_id = "10543"
                >> version = JIRA("https://atlassian.org").version(version_id)
                >> print(version.name)
                "some_version_name"
                >> version.update(name="another_name")
                >> print(version.name)
                "another_name"
                >> version.update(archived=True)
                >> print(version.archived)
                True
        """
        data = {}
        for field in kwargs:
            data[field] = kwargs[field]

        super().update(**data)


# GreenHopper


class GreenHopperResource(Resource):
    """A generic GreenHopper resource."""

    AGILE_BASE_URL = "{server}/rest/{agile_rest_path}/{agile_rest_api_version}/{path}"

    GREENHOPPER_REST_PATH = "greenhopper"
    """ Old, private API. Deprecated and will be removed from Jira on the 1st February 2016. """
    AGILE_EXPERIMENTAL_REST_PATH = "greenhopper/experimental-api"
    """ Experimental API available in Jira Agile 6.7.3 - 6.7.6, basically the same as Public API """
    AGILE_BASE_REST_PATH = "agile"
    """ Public API introduced in Jira Agile 6.7.7. """

    def __init__(
        self,
        path: str,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        self.self = None

        Resource.__init__(self, path, options, session, self.AGILE_BASE_URL)
        if raw:
            self._parse_raw(raw)
            # Old GreenHopper API did not contain self - create it for backward compatibility.
            if not self.self:
                self.self = self._get_url(path.format(raw["id"]))
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class Sprint(GreenHopperResource):
    """A GreenHopper sprint."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        GreenHopperResource.__init__(self, "sprint/{0}", options, session, raw)

    def find(self, id, params=None):
        if (
            self._options["agile_rest_path"]
            != GreenHopperResource.GREENHOPPER_REST_PATH
        ):
            Resource.find(self, id, params)
        else:
            # Old, private GreenHopper API had non-standard way of loading Sprint
            url = self._get_url(f"sprint/{id}/edit/model")
            self._load(url, params=params, path="sprint")


class Board(GreenHopperResource):
    """A GreenHopper board."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        path = (
            "rapidview/{0}"
            if options["agile_rest_path"] == self.GREENHOPPER_REST_PATH
            else "board/{id}"
        )
        GreenHopperResource.__init__(self, path, options, session, raw)

    def delete(self, params=None):
        if (
            self._options["agile_rest_path"]
            != GreenHopperResource.GREENHOPPER_REST_PATH
        ):
            raise NotImplementedError(
                "Jira Agile Public API does not support Board removal"
            )

        Resource.delete(self, params)


# Service Desk


class Customer(Resource):
    """A Service Desk customer."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(
            self, "customer", options, session, "{server}/rest/servicedeskapi/{path}"
        )
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class ServiceDesk(Resource):
    """A Service Desk."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(
            self,
            "servicedesk/{0}",
            options,
            session,
            "{server}/rest/servicedeskapi/{path}",
        )
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


class RequestType(Resource):
    """A Service Desk Request Type."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(
            self,
            "servicedesk/{0}/requesttype",
            options,
            session,
            "{server}/rest/servicedeskapi/{path}",
        )

        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


# Utilities


def dict2resource(
    raw: Dict[str, Any], top=None, options=None, session=None
) -> Union["PropertyHolder", Type[Resource]]:
    """Convert a dictionary into a Jira Resource object.

    Recursively walks a dict structure, transforming the properties into attributes
    on a new ``Resource`` object of the appropriate type (if a ``self`` link is present)
    or a ``PropertyHolder`` object (if no ``self`` link is present).
    """
    if top is None:
        top = PropertyHolder(raw)

    seqs = tuple, list, set, frozenset
    for i, j in raw.items():
        if isinstance(j, dict):
            if "self" in j:
                # to try and help mypy know that cls_for_resource can never be 'Resource'
                resource_class = cast(Type[Resource], cls_for_resource(j["self"]))
                resource = cast(
                    Type[Resource],
                    resource_class(  # type: ignore
                        options=options, session=session, raw=j  # type: ignore
                    ),
                )
                setattr(top, i, resource)
            elif i == "timetracking":
                setattr(top, "timetracking", TimeTracking(options, session, j))
            else:
                setattr(top, i, dict2resource(j, options=options, session=session))
        elif isinstance(j, seqs):
            j = cast(List[Dict[str, Any]], j)  # help mypy
            seq_list: List[Any] = []
            for seq_elem in j:
                if isinstance(seq_elem, dict):
                    if "self" in seq_elem:
                        # to try and help mypy know that cls_for_resource can never be 'Resource'
                        resource_class = cast(
                            Type[Resource], cls_for_resource(seq_elem["self"])
                        )
                        resource = cast(
                            Type[Resource],
                            resource_class(  # type: ignore
                                options=options,
                                session=session,
                                raw=seq_elem,  # type: ignore
                            ),
                        )
                        seq_list.append(resource)
                    else:
                        seq_list.append(
                            dict2resource(seq_elem, options=options, session=session)
                        )
                else:
                    seq_list.append(seq_elem)
            setattr(top, i, seq_list)
        else:
            setattr(top, i, j)
    return top


resource_class_map: Dict[str, Type[Resource]] = {
    # Jira-specific resources
    r"attachment/[^/]+$": Attachment,
    r"component/[^/]+$": Component,
    r"customFieldOption/[^/]+$": CustomFieldOption,
    r"dashboard/[^/]+$": Dashboard,
    r"filter/[^/]$": Filter,
    r"issue/[^/]+$": Issue,
    r"issue/[^/]+/comment/[^/]+$": Comment,
    r"issue/[^/]+/votes$": Votes,
    r"issue/[^/]+/watchers$": Watchers,
    r"issue/[^/]+/worklog/[^/]+$": Worklog,
    r"issueLink/[^/]+$": IssueLink,
    r"issueLinkType/[^/]+$": IssueLinkType,
    r"issuetype/[^/]+$": IssueType,
    r"priority/[^/]+$": Priority,
    r"project/[^/]+$": Project,
    r"project/[^/]+/role/[^/]+$": Role,
    r"project/[^/]+/permissionscheme[^/]+$": PermissionScheme,
    r"resolution/[^/]+$": Resolution,
    r"securitylevel/[^/]+$": SecurityLevel,
    r"status/[^/]+$": Status,
    r"statuscategory/[^/]+$": StatusCategory,
    r"user\?(username|key|accountId).+$": User,
    r"group\?groupname.+$": Group,
    r"version/[^/]+$": Version,
    # GreenHopper specific resources
    r"sprints/[^/]+$": Sprint,
    r"views/[^/]+$": Board,
}


class UnknownResource(Resource):
    """A Resource from Jira that is not (yet) supported."""

    def __init__(
        self,
        options: Dict[str, str],
        session: ResilientSession,
        raw: Dict[str, Any] = None,
    ):
        Resource.__init__(self, "unknown{0}", options, session)
        if raw:
            self._parse_raw(raw)
        self.raw: Dict[str, Any] = cast(Dict[str, Any], self.raw)


def cls_for_resource(resource_literal: str) -> Type[Resource]:
    for resource in resource_class_map:
        if re.search(resource, resource_literal):
            return resource_class_map[resource]
    else:
        # Generic Resource cannot directly be used b/c of different constructor signature
        return UnknownResource


class PropertyHolder:
    def __init__(self, raw):
        __bases__ = raw  # noqa



#!/usr/bin/env python
#
# This script will cleanup your jira instance by removing all projects and
# it is used to clean the CI/CD Jira server used for testing.
#
import json
import logging
import os

from jira import JIRA, Issue, JIRAError, Project, Role  # noqa

logging.getLogger().setLevel(logging.DEBUG)
logging.getLogger("requests").setLevel(logging.INFO)
logging.getLogger("urllib3").setLevel(logging.INFO)
logging.getLogger("jira").setLevel(logging.DEBUG)

CI_JIRA_URL = os.environ["CI_JIRA_URL"]
CI_JIRA_ADMIN = os.environ["CI_JIRA_ADMIN"]
CI_JIRA_ADMIN_PASSWORD = os.environ["CI_JIRA_ADMIN_PASSWORD"]

j = JIRA(
    CI_JIRA_URL,
    basic_auth=(CI_JIRA_ADMIN, CI_JIRA_ADMIN_PASSWORD),
    logging=True,
    validate=True,
    async_=True,
    async_workers=20,
)

logging.info("Running maintenance as %s", j.current_user())

for p in j.projects():
    logging.info("Deleting project %s", p)
    try:
        j.delete_project(p)
    except Exception as e:
        logging.error(e)

for s in j.permissionschemes():
    if " for Project" in s["name"]:
        logging.info("Deleting permission scheme: %s" % s["name"])
        try:
            j.delete_permissionscheme(s["id"])
        except JIRAError as e:
            logging.error(e.text)
    else:
        logging.info("Permission scheme: %s" % s["name"])

for s in j.issuesecurityschemes():
    if " for Project" in s["name"]:
        logging.info("Deleting issue security scheme: %s", s["name"])
        j.delete_permissionscheme(s["id"])
    else:
        logging.error("Issue security scheme: %s" % s["name"])

for s in j.projectcategories():
    # if ' for Project' in s['name']:
    #     print("Deleting issue security scheme: %s" % s['name'])
    #     # j.delete_permissionscheme(s['id'])
    # else:
    logging.info("Project category: %s" % s["name"])

for s in j.avatars("project"):
    logging.info("Avatar project: %s", s)

# disabled until Atlassian implements DELETE verb
# for s in j.screens():
#     if s['id'] >= 1000:
#         try:
#             logging.info("Deleting screen: %s" % s['name'])
#             j.delete_screen(s['id'])
#         except Exception as e:
#             logging.error(e)
#     else:
#         logging.error(s)

for s in j.notificationschemes():
    logging.info("NotificationScheme: %s", s)

# TODO(ssbarnea): "Default Issue Security Scheme"

for t in j.templates():
    logging.info("ProjectTemplate: %s", json.dumps(t, indent=4, sort_keys=True))



"""Jira utils used internally."""
import threading
import warnings
from typing import Any, Optional, cast

from requests import Response
from requests.structures import CaseInsensitiveDict as _CaseInsensitiveDict

from jira.resilientsession import raise_on_error


class CaseInsensitiveDict(_CaseInsensitiveDict):
    """A case-insensitive ``dict``-like object.

    DEPRECATED: use requests.structures.CaseInsensitiveDict directly.

    Implements all methods and operations of
    ``collections.MutableMapping`` as well as dict's ``copy``. Also
    provides ``lower_items``.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``
    will contain case-sensitive keys. However, querying and contains
    testing is case insensitive::

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['accept'] == 'application/json'  # True
        list(cid) == ['Accept']  # True

    For example, ``headers['content-encoding']`` will return the
    value of a ``'Content-Encoding'`` response header, regardless
    of how the header name was originally stored.

    If the constructor, ``.update``, or equality comparison
    operations are given keys that have equal ``.lower()`` s, the
    behavior is undefined.

    """

    def __init__(self, *args, **kwargs) -> None:
        warnings.warn(
            "Use requests.structures.CaseInsensitiveDict directly", DeprecationWarning
        )
        super().__init__(*args, **kwargs)


def threaded_requests(requests):
    for fn, url, request_args in requests:
        th = threading.Thread(target=fn, args=(url,), kwargs=request_args, name=url)
        th.start()

    for th in threading.enumerate():
        if th.name.startswith("http"):
            th.join()


def json_loads(r: Optional[Response]) -> Any:
    """Attempts to load json the result of a response

    Args:
        r (Optional[Response]): The Response object

    Raises:
        JIRAError: via :py:func:`jira.resilientsession.raise_on_error`

    Returns:
        Union[List[Dict[str, Any]], Dict[str, Any]]: the json
    """
    raise_on_error(r)  # if 'r' is None, will raise an error here
    r = cast(Response, r)  # tell mypy only Response-like are here
    try:
        return r.json()
    except ValueError:
        # json.loads() fails with empty bodies
        if not r.text:
            return {}
        raise



"""The root of JIRA package namespace."""
try:
    import pkg_resources

    __version__ = pkg_resources.get_distribution("jira").version
except Exception:
    __version__ = "unknown"

from jira.client import JIRA  # noqa: E402
from jira.client import Comment  # noqa: E402
from jira.client import Issue  # noqa: E402
from jira.client import Priority  # noqa: E402
from jira.client import Project  # noqa: E402
from jira.client import Role  # noqa: E402
from jira.client import User  # noqa: E402
from jira.client import Watchers  # noqa: E402
from jira.client import Worklog  # noqa: E402
from jira.config import get_jira  # noqa: E402
from jira.exceptions import JIRAError  # noqa: E402

__all__ = (
    "Comment",
    "__version__",
    "Issue",
    "JIRA",
    "JIRAError",
    "Priority",
    "Project",
    "Role",
    "User",
    "Watchers",
    "Worklog",
    "get_jira",
)
